use clap::{crate_authors, crate_description, crate_version, ArgMatches};
use clap::{App, Arg, SubCommand};
use client::archive::Archive;
use dotenv::dotenv;
use log::{debug, info};
use save_sync::db::{establish_connection, Database};
use std::path::Path;

fn main() {
    dotenv().ok();
    env_logger::init();

    let app = App::new("Save Sync")
        .version(crate_version!())
        .author(crate_authors!())
        .about(crate_description!());
    let m = app
        .subcommand(
            SubCommand::with_name("track")
                .arg(
                    Arg::with_name("path")
                        .value_name("PATH")
                        .takes_value(true)
                        .required(true)
                        .index(1)
                        .help("The file / directory which will be tracked"),
                )
                .arg(
                    Arg::with_name("friendly")
                        .short("f")
                        .long("friendly")
                        .value_name("NAME")
                        .takes_value(true)
                        .help("A friendly name for a tracked file / directory"),
                ),
        )
        .subcommand(
            SubCommand::with_name("info")
                .arg(
                    Arg::with_name("path")
                        .value_name("PATH")
                        .takes_value(true)
                        .required_unless("friendly")
                        .index(1)
                        .help("The path of the game save"),
                )
                .arg(
                    Arg::with_name("friendly")
                        .long("friendly")
                        .short("f")
                        .value_name("NAME")
                        .takes_value(true)
                        .help("The friendly name of the game save"),
                ),
        )
        .subcommand(SubCommand::with_name("list"))
        .subcommand(
            SubCommand::with_name("drop")
                .arg(
                    Arg::with_name("path")
                        .value_name("PATH")
                        .takes_value(true)
                        .required_unless("friendly")
                        .index(1)
                        .help("The path of the files you no longer want to track"),
                )
                .arg(
                    Arg::with_name("friendly")
                        .long("friendly")
                        .short("f")
                        .value_name("NAME")
                        .takes_value(true)
                        .help("The friendly name of the game save you no longer want tracked."),
                ),
        )
        .get_matches();

    match m.subcommand() {
        ("track", Some(sub_m)) => track_save(sub_m),
        ("drop", Some(sub_m)) => drop_save(sub_m),
        ("info", Some(sub_m)) => tracked_save_info(sub_m),
        ("list", Some(sub_m)) => list_tracked_saves(sub_m),
        _ => eprintln!("No valid subcommand / argument provided"),
    }
}

fn track_save(matches: &ArgMatches) {
    let path_str = matches
        .value_of("path")
        .expect("The path argument was not set despite it being required");
    let mut archive = Archive::try_default().expect("Failed to create a new Archive");

    if let Some(f_name) = matches.value_of("friendly") {
        info!("Name {} present for {:?}", f_name, path_str);
        archive
            .track_game_with_friendly(path_str, f_name)
            .expect("Archive failed to track Game Save Location")
    } else {
        info!("No friendly name present for {:?}", path_str);
        archive
            .track_game(path_str)
            .expect("Archive failed to track Game Save Location");
    }

    info!("Now Tracking: {:?}", path_str);
}

fn tracked_save_info(matches: &ArgMatches) {
    let maybe_game = if let Some(f_name) = matches.value_of("friendly") {
        Archive::get_game_with_friendly(f_name).expect("Failed to get game save info from Archive")
    } else {
        // There is guaranteed to be a path given by the user
        let path_str = matches
            .value_of("path")
            .expect("The path argument was not set despite it being required");
        Archive::get_game(path_str).expect("Failed to get game save info from Archive")
    };

    let game = maybe_game.expect("No tracked game save found");

    if let Some(name) = game.friendly_name {
        println!("Friendly Name: {}", name);
    } else {
        println!("Friendly Name: None");
    }

    println!("Original Path: {:?}", game.original_path);
    println!("UUID: {:?}", game.uuid);
    println!("---\nFiles:");

    for file in game.files {
        println!("Path: {:?}", file.original_path);
        println!("Hash: {}", file.hash);
        println!();
    }
}

fn list_tracked_saves(_matches: &ArgMatches) {
    let games = Archive::get_all_games()
        .expect("Failed to get all Games from the Archive")
        .expect("There are no tracked Games");

    for game in games {
        if let Some(name) = game.friendly_name {
            print!("[{}] ", name);
        }
        println!("{:?}", game.original_path);
        println!("UUID: {:?}", game.uuid);
        println!("---");
    }
}

fn drop_save(matches: &ArgMatches) {
    let maybe_compromised = if let Some(name) = matches.value_of("friendly") {
        Archive::drop_game_with_friendly(name).expect("Archive failed to delete from database")
    } else {
        let path_str = matches
            .value_of("path")
            .expect("The path argument was not set despite it being required");

        Archive::drop_game(path_str).expect("Archive failed to delete from database")
    };

    match maybe_compromised {
        Some(()) => println!("Game successfully dropped from the list"),
        None => panic!("Database Invariant broken. Database is corrupted."),
    }
}
